# -*- coding: utf-8 -*-
# Quasarr
# Project by https://github.com/rix1337

def render_centered_html(inner_content):
    style_outer = """
    display: flex;
    justify-content: center;
    align-items: center;
    height: 100vh;
    max-height: 100vh;
    overflow-y: auto;
    overflow-x: hidden; /* Prevent horizontal scrolling */
    width: 100%; /* Ensure it spans full width */
    background-color: #212529;
    color: #fff;
    font-family: system-ui,-apple-system,'Segoe UI',Roboto,'Helvetica Neue',
    'Noto Sans','Liberation Sans',Arial,sans-serif,'Apple Color Emoji',
    'Segoe UI Emoji','Segoe UI Symbol','Noto Color Emoji';
    """
    style_inner = """
    background-color: #fff;
    border-radius: 0.375rem;
    box-shadow: 0 0.5rem 1rem rgba(0, 0, 0, 0.15);
    padding: 4px;
    text-align: center;
    color: #212529;
    font-size: 1rem;
    font-weight: 400;
    line-height: 1.5;
    box-sizing: border-box; /* Ensure padding doesn’t exceed boundaries */
    max-width: 100%; /* Allow content to shrink */
    margin: auto;
    """

    return f'''
    <html>
    <head>
        <meta name="viewport" content="width=device-width, initial-scale=1">
        <title>Quasarr</title>
    </head>
    <body>
    <div style="{style_outer.strip()}">
        <div style="{style_inner.strip()}">
            {inner_content}
        </div>
    </div>
    </body>
    </html>
    '''


def render_button(text, button_type="primary", attributes=None):
    base_style = (
        "padding: 0.375rem 0.75rem; font-size: 1rem; line-height: 1.5; "
        "border-radius: 0.375rem; color: #fff; display: inline-block; "
        "font-weight: 400; text-align: center; vertical-align: middle; "
        "cursor: pointer; -webkit-user-select: none; -moz-user-select: none; "
        "user-select: none; transition: color 0.15s ease-in-out, "
        "background-color 0.15s ease-in-out, border-color 0.15s ease-in-out, "
        "box-shadow 0.15s ease-in-out; "
    )

    if button_type == "primary":
        style = base_style + "background-color: #0d6efd; border: 1px solid #0d6efd; "
    else:
        style = base_style + "background-color: #6c757d; border: 1px solid #6c757d; "

    attr_str = ' '.join(f'{key}="{value}"' for key, value in attributes.items()) if attributes else ""

    return f'<button style="{style}" {attr_str}>{text}</button>'


def render_form(header, form="", script=""):
    styles = """
    <style>
        input, select {
            display: block;
            padding: .375rem .75rem;
            width: 100%;
            font-size: 1rem;
            font-weight: 400;
            line-height: 1.5;
            color: #212529;
            background-color: #fff;
            border: 1px solid #dee2e6;
            border-radius: .375rem;
            transition: border-color .15s ease-in-out, box-shadow .15s ease-in-out;
            text-align: center;
            margin: 10px auto;
        }
    </style>
    """
    content = f'''
    <h1>Quasarr</h1>
    <h3>{header}</h3>
    {styles}
    {form}
    {script}
    '''
    return render_centered_html(content)


def render_success(message, timeout=10):
    button_html = render_button(f"Wait time... {timeout}", "secondary", {"id": "nextButton", "disabled": "true"})
    script = f"""
                <script>
                    var counter = {timeout};
                    var interval = setInterval(function() {{
                        counter--;
                        document.getElementById('nextButton').innerText = 'Wait time... ' + counter;
                        if (counter === 0) {{
                            clearInterval(interval);
                            document.getElementById('nextButton').innerText = 'Continue';
                            document.getElementById('nextButton').disabled = false;
                            document.getElementById('nextButton').onclick = function() {{
                                window.location.href='/';
                            }};
                            // Change button style to primary
                            document.getElementById('nextButton').style.backgroundColor = '#0d6efd';
                            document.getElementById('nextButton').style.borderColor = '#0d6efd';
                        }}
                    }}, 1000);
                </script>
                """
    content = f"""<h1>Quasarr</h1>
                <h3>{message}</h3>
                {button_html}
                {script}
                """
    return render_centered_html(content)


def render_fail(message):
    button_html = render_button("Back", "secondary", {"onclick": "window.location.href='/'"})
    return render_centered_html(f"""<h1>Quasarr</h1>
                        <h3>{message}</h3>
                        {button_html}
                        """)



# -*- coding: utf-8 -*-
# Quasarr
# Project by https://github.com/rix1337

import json
import re
import traceback
import xml.sax.saxutils as sax_utils
from base64 import urlsafe_b64decode
from datetime import datetime
from xml.etree import ElementTree as element_tree

import requests
from bottle import Bottle, request, response, abort

from quasarr.captcha_solver import get_filecrypt_links
from quasarr.downloads import download_package, delete_package, get_packages
from quasarr.providers import shared_state
from quasarr.providers.html_templates import render_button, render_centered_html
from quasarr.providers.notifications import send_discord_message
from quasarr.providers.obfuscated import captcha_js, captcha_values
from quasarr.providers.tvmaze_metadata import get_title_from_tvrage_id
from quasarr.providers.web_server import Server
from quasarr.search import get_search_results


def api(shared_state_dict, shared_state_lock):
    shared_state.set_state(shared_state_dict, shared_state_lock)

    app = Bottle()

    @app.get('/captcha')
    def serve_captcha():
        try:
            device = shared_state.values["device"]
        except KeyError:
            device = None
        if not device:
            return render_centered_html(f'''<h1>Quasarr</h1>
            <p>JDownloader connection not established.</p>''')

        protected = shared_state.get_db("protected").retrieve_all_titles()
        if not protected:
            return render_centered_html(f'''<h1>Quasarr</h1>
            <p>No protected packages found! CAPTCHA not needed.</p>''')
        else:
            package = protected[0]
            package_id = package[0]
            data = json.loads(package[1])
            title = data["title"]
            links = data["links"]
            password = data["password"]

        link_options = ""
        if len(links) > 1:
            for link in links:
                if "filecrypt." in link[0]:
                    link_options += f'<option value="{link[0]}">{link[1]}</option>'
            link_select = f'''<div id="mirrors-select">
                    <label for="link-select">Mirror:</label>
                    <select id="link-select">
                        {link_options}
                    </select>
                </div>
                <script>
                    document.getElementById("link-select").addEventListener("change", function() {{
                        var selectedLink = this.value;
                        document.getElementById("link-hidden").value = selectedLink;
                    }});
                </script>
            '''
        else:
            link_select = f'<div id="mirrors-select">Mirror: <b>{links[0][1]}</b></div>'

        content = render_centered_html(r'''
            <script type="text/javascript">
                var api_key = "''' + captcha_values()["api_key"] + r'''";
                var endpoint = '/' + window.location.pathname.split('/')[1] + '/' + api_key + '.html';
                function handleToken(token) {
                    document.getElementById("puzzle-captcha").remove();
                    document.getElementById("mirrors-select").remove();
                    document.getElementById("captcha-key").innerText = 'Using result "' + token + '" to decrypt links...';
                    var link = document.getElementById("link-hidden").value;
                    const currentPath = window.location.pathname;
                    const endpoint = '/decrypt-filecrypt';
                    const fullPath = currentPath.endsWith('/') ? currentPath + endpoint.slice(1) : currentPath + endpoint;

                    fetch(fullPath, {
                        method: 'POST',
                        headers: {
                            'Content-Type': 'application/json',
                        },
                        body: JSON.stringify({ 
                            token: token,
                            ''' + f'''package_id: '{package_id}',
                            title: '{title}',
                            link: link,
                            password: '{password}'
                        ''' + '''})
                    })
                    .then(response => response.json())
                    .then(data => {
                        if (data.success) {
                            document.getElementById("captcha-key").insertAdjacentHTML('afterend', 
                                '<p>Successful for: ' + data.title + '</p>');
                        } else {
                            document.getElementById("captcha-key").insertAdjacentHTML('afterend', 
                                '<p>Failed. Check console for details!</p>');
                        }
                        document.getElementById("reload-button").style.display = "block";
                    });
                }
                ''' + captcha_js() + f'''</script>
                <div>
                    <h1>Quasarr</h1>
                    <div id="captcha-key"></div>
                    {link_select}<br><br>
                    <input type="hidden" id="link-hidden" value="{links[0][0]}" />
                    <div id="puzzle-captcha" aria-style="mobile">
                        <strong>Your adblocker prevents the captcha from loading. Disable it!</strong>
                    </div>
                    <div id="reload-button" style="display: none;">
                    {render_button("Solve another CAPTCHA", "secondary", {
            "onclick": "location.reload()",
        })}</div>
                </div>
                </html>''')

        return content

    @app.post('/captcha/decrypt-filecrypt')
    def submit_token():
        protected = shared_state.get_db("protected").retrieve_all_titles()
        if not protected:
            return {"success": False, "title": "No protected packages found! CAPTCHA not needed."}

        download_links = []

        try:
            data = request.json
            token = data.get('token')
            package_id = data.get('package_id')
            title = data.get('title')
            link = data.get('link')
            password = data.get('password')
            if token:
                print(f"Received token: {token}")
                print(f"Decrypting links for {title}")
                download_links = get_filecrypt_links(shared_state, token, title, link, password)

                print(f"Decrypted {len(download_links)} download links for {title}")

                shared_state.download_package(download_links, title, password, package_id)

                shared_state.get_db("protected").delete(package_id)

        except Exception as e:
            print(f"Error decrypting: {e}")

        return {"success": bool(download_links), "title": title}

    @app.post('/captcha/<captcha_id>.html')
    def proxy(captcha_id):
        target_url = f"{captcha_values()["url"]}/captcha/{captcha_id}.html"

        headers = {key: value for key, value in request.headers.items() if key != 'Host'}
        data = request.body.read()
        resp = requests.post(target_url, headers=headers, data=data)

        response.content_type = resp.headers.get('Content-Type')

        content = resp.text
        content = re.sub(r'<script src="/(.*?)"></script>',
                         f'<script src="{captcha_values()["url"]}/\\1"></script>', content)
        response.content_type = 'text/html'
        return content

    @app.post('/captcha/<captcha_id>.json')
    def specific_proxy(captcha_id):
        target_url = f"{captcha_values()["url"]}/captcha/{captcha_id}.json"

        headers = {key: value for key, value in request.headers.items() if key != 'Host'}
        data = request.body.read()
        resp = requests.post(target_url, headers=headers, data=data)

        response.content_type = resp.headers.get('Content-Type')
        return resp.content

    @app.get('/captcha/<captcha_id>/<uuid>/<filename>')
    def captcha_proxy(captcha_id, uuid, filename):
        new_url = f"{captcha_values()["url"]}/captcha/{captcha_id}/{uuid}/{filename}"

        try:
            external_response = requests.get(new_url, stream=True)
            external_response.raise_for_status()
            response.content_type = 'image/png'
            response.headers['Content-Disposition'] = f'inline; filename="{filename}"'
            return external_response.iter_content(chunk_size=8192)

        except requests.RequestException as e:
            response.status = 502
            return f"Error fetching resource: {e}"

    @app.post('/captcha/<captcha_id>/check')
    def captcha_check_proxy(captcha_id):
        new_url = f"{captcha_values()["url"]}/captcha/{captcha_id}/check"
        headers = {key: value for key, value in request.headers.items()}

        data = request.body.read()
        resp = requests.post(new_url, headers=headers, data=data)

        response.status = resp.status_code
        for header in resp.headers:
            if header.lower() not in ['content-encoding', 'transfer-encoding', 'content-length', 'connection']:
                response.set_header(header, resp.headers[header])
        return resp.content

    @app.get('/')
    def index():
        protected = shared_state.get_db("protected").retrieve_all_titles()
        captcha_hint = ""
        if protected:
            package_count = len(protected)
            package_text = f"Package{'s' if package_count > 1 else ''} waiting for CAPTCHA"
            amount_info = f": {package_count}" if package_count > 1 else ""
            button_text = f"Solve CAPTCHA{'s' if package_count > 1 else ''} here to decrypt links!"

            captcha_hint = f'''
            <h2>Links protected by CAPTCHA</h2>
            <p>{package_text}{amount_info}</p>
            <p>{render_button(button_text, "primary", {"onclick": "location.href='/captcha'"})}</p>
            '''

        info = f"""
        <h1>Quasarr</h1>
        <p>
            <code id="current-url" style="background-color: #f0f0f0; padding: 5px; border-radius: 3px;">
                {shared_state.values["internal_address"]}
            </code>
        </p>
        <p>Use this exact URL as 'Newznab Indexer' and 'SABnzbd Download Client' in Sonarr/Radarr.<br>
        Leave settings at default and use this API key: 'quasarr'</p>
        {captcha_hint}
        """
        return render_centered_html(info)

    @app.get('/download/')
    def fake_download_container():
        payload = request.query.payload
        decoded_payload = urlsafe_b64decode(payload).decode("utf-8").split("|")
        title = decoded_payload[0]
        url = decoded_payload[1]
        size_mb = decoded_payload[2]
        password = decoded_payload[3]
        return f'<nzb><file title="{title}" url="{url}" size_mb="{size_mb}" password="{password}"/></nzb>'

    @app.post('/api')
    def download():
        downloads = request.files.getall('name')
        nzo_ids = []
        for upload in downloads:
            file_content = upload.file.read()
            root = element_tree.fromstring(file_content)
            title = sax_utils.unescape(root.find(".//file").attrib["title"])
            url = root.find(".//file").attrib["url"]
            size_mb = root.find(".//file").attrib["size_mb"]
            password = root.find(".//file").attrib.get("password")
            print(f"Attempting download for {title}")

            request_from = request.headers.get('User-Agent')

            nzo_id = download_package(shared_state, request_from, title, url, size_mb, password)
            if nzo_id:
                print(f"{title} added successfully!")
                nzo_ids.append(nzo_id)
            else:
                print(f"{title} could not be added!")

        return {
            "status": True,
            "nzo_ids": nzo_ids
        }

    @app.get('/api')
    def fake_api():
        api_type = 'sabnzbd' if request.query.mode and request.query.apikey else 'newznab' if request.query.t else None

        if api_type == 'sabnzbd':
            try:
                mode = request.query.mode
                if mode == "version":
                    return {
                        "version": "4.3.2"
                    }
                elif mode == "get_config":
                    return {
                        "config": {
                            "misc": {
                                "quasarr": True,
                                "complete_dir": "/tmp/"
                            },
                            "categories": [
                                {
                                    "name": "*",
                                    "order": 0,
                                    "dir": "",
                                },
                                {
                                    "name": "movies",
                                    "order": 1,
                                    "dir": "",
                                },
                                {
                                    "name": "tv",
                                    "order": 2,
                                    "dir": "",
                                }
                            ]
                        }
                    }
                elif mode == "fullstatus":
                    return {
                        "status": {
                            "quasarr": True
                        }
                    }
                elif mode == "queue" or mode == "history":
                    if request.query.name and request.query.name == "delete":
                        package_id = request.query.value
                        deleted = delete_package(shared_state, package_id)
                        return {
                            "status": deleted,
                            "nzo_ids": [package_id]
                        }

                    packages = get_packages(shared_state)
                    if mode == "queue":
                        return {
                            "queue": {
                                "paused": False,
                                "slots": packages["queue"]
                            }
                        }
                    elif mode == "history":
                        return {
                            "history": {
                                "paused": False,
                                "slots": packages["history"]
                            }
                        }
            except Exception as e:
                print(f"Error loading packages: {e}")
                print(traceback.format_exc())
            return {
                "status": False
            }

        elif api_type == 'newznab':
            try:
                mode = request.query.t
                if mode == 'caps':
                    return '''<?xml version="1.0" encoding="UTF-8"?>
                                    <caps>
                                      <categories>
                                          <category id="2000" name="Movies">
                                              <subcat id="2010" name="Foreign"/>
                                              <subcat id="2020" name="Other"/>
                                              <subcat id="2030" name="SD"/>
                                              <subcat id="2040" name="HD"/>
                                              <subcat id="2050" name="BluRay"/>
                                              <subcat id="2060" name="3D"/>
                                          </category>
                                          <category id="5000" name="TV">
                                              <subcat id="5020" name="Foreign"/>
                                              <subcat id="5030" name="SD"/>
                                              <subcat id="5040" name="HD"/>
                                              <subcat id="5050" name="Other"/>
                                              <subcat id="5060" name="Sport"/>
                                          </category>
                                      </categories>
                                    </caps>'''
                elif mode in ['movie', 'tvsearch']:
                    request_from = request.headers.get('User-Agent')

                    releases = []

                    if mode == 'movie':
                        # only imdb is implemented
                        search_param = f"tt{getattr(request.query, 'imdbid', '')}" \
                            if getattr(request.query, 'imdbid', '') else ""

                        releases = get_search_results(shared_state, request_from, search_string=search_param)

                    elif mode == 'tvsearch':
                        # these are currently ignored, Sonarr handles them anyway
                        season = getattr(request.query, 'season', "")
                        episode = getattr(request.query, 'ep', "")
                        # only plain search string and tvrage id is implemented
                        search_param = getattr(request.query, 'q', "")
                        if not search_param:
                            tvrage_id = getattr(request.query, 'rid', "")
                            if tvrage_id:
                                search_param = get_title_from_tvrage_id(tvrage_id)

                        offset = getattr(request.query, 'offset', "")  # ignoring offset higher than 0 on purpose
                        if int(offset) == 0:
                            releases = get_search_results(shared_state, request_from,
                                                          search_string=search_param,
                                                          season=season,
                                                          episode=episode
                                                          )

                    items = ""
                    if not releases:
                        items += f'''
                            <item>
                                <title>No releases found</title>
                                <link></link>
                                <pubDate>{datetime.now().strftime('%a, %d %b %Y %H:%M:%S %z')}</pubDate>
                                <enclosure url="_" length="0" type="application/x-nzb"/>
                                <guid></guid>
                                <comments></comments>
                                <description></description>
                            </item>'''

                    for release in releases:
                        release = release["details"]
                        items += f'''
                        <item>
                            <title>{sax_utils.escape(release["title"])}</title>
                            <guid isPermaLink="True">{release["link"]}</guid>
                            <link>{release["link"]}</link>
                            <comments>{release["source"]}</comments>
                            <pubDate>{release["date"]}</pubDate>
                            <enclosure url="{release["link"]}" length="{release["size"]}" type="application/x-nzb" />
                        </item>'''

                    return f'''<?xml version="1.0" encoding="UTF-8"?>
                                <rss version="2.0">
                                    <channel>
                                        {items}
                                    </channel>
                                </rss>'''
            except Exception as e:
                print(f"Error loading search results: {e}")
                print(traceback.format_exc())

            print(f"Unknown request: {request.query}")
            return {"error": True}

    @app.get("/sponsors_helper/api/to_decrypt/")
    def to_decrypt_api():
        global helper_active
        try:
            helper_active = True

            protected = shared_state.get_db("protected").retrieve_all_titles()
            if not protected:
                return abort(404, "No encrypted packages found")
            else:
                package = protected[0]
                package_id = package[0]
                data = json.loads(package[1])
                title = data["title"]
                links = data["links"]
                password = data["password"]

            return {
                "to_decrypt": {
                    "name": title,
                    "id": package_id,
                    "url": links,
                    "password": password,
                    "max_attempts": 3
                }
            }
        except:
            return abort(500, "Failed")

    @app.post("/sponsors_helper/api/to_download/")
    def to_download_api():
        try:
            data = request.json
            title = data.get('name')
            package_id = data.get('package_id')
            download_links = data.get('urls')
            password = data.get('password')

            print(f"Received {len(download_links)} download links for {title}")

            if download_links:
                downloaded = shared_state.download_package(download_links, title, password, package_id)
                if downloaded:
                    shared_state.get_db("protected").delete(package_id)
                    send_discord_message(shared_state, title=title, case="solved")
                    print(f"Download successfully started for {title}")
                    return f"Downloaded {len(download_links)} download links for {title}"
                else:
                    print(f"Download failed for {title}")

        except Exception as e:
            print(f"Error decrypting: {e}")

        return abort(500, "Failed")

    @app.delete("/sponsors_helper/api/to_delete/")
    def to_delete_api():
        try:
            data = request.json
            package_id = data.get('package_id')
            deleted = delete_package(shared_state, package_id)
            if deleted:
                send_discord_message(shared_state, title=deleted, case="deleted")
                return f"Deleted package {deleted} with id {package_id}"

        except Exception as e:
            print(f"Error deleting: {e}")

        return abort(500, "Failed")

    @app.put("/sponsors_helper/api/activate_sponsor_status/")
    def activate_sponsor_status():
        try:
            data = request.body.read().decode("utf-8")
            payload = json.loads(data)
            if payload["activate"]:
                shared_state.update("helper_active", True)
                print(f"Sponsor status activated successfully")
                return "Sponsor status activated successfully!"
        except:
            pass
        return abort(500, "Failed")

    Server(app, listen='0.0.0.0', port=shared_state.values["port"]).serve_forever()



# -*- coding: utf-8 -*-
# Quasarr
# Project by https://github.com/rix1337

import argparse
import multiprocessing
import os
import re
import socket
import sys
import tempfile
import time
from urllib.parse import urlparse

import requests

from quasarr.arr import api
from quasarr.providers import shared_state, version
from quasarr.storage.config import Config, get_clean_hostnames
from quasarr.storage.setup import path_config, hostnames_config, nx_credentials_config, jdownloader_config
from quasarr.storage.sqlite_database import DataBase


def run():
    with multiprocessing.Manager() as manager:
        shared_state_dict = manager.dict()
        shared_state_lock = manager.Lock()
        shared_state.set_state(shared_state_dict, shared_state_lock)

        parser = argparse.ArgumentParser()
        parser.add_argument("--port", help="Desired Port, defaults to 8080")
        parser.add_argument("--internal_address", help="Must be provided when running in Docker")
        parser.add_argument("--external_address", help="External address for CAPTCHA notifications")
        parser.add_argument("--discord", help="Discord Webhook URL")
        parser.add_argument("--hostnames", help="Public HTTP(s) Link that contains hostnames definition.")
        arguments = parser.parse_args()

        sys.stdout = Unbuffered(sys.stdout)

        print(f"""┌────────────────────────────────────┐
  Quasarr {version.get_version()} by RiX
  https://github.com/rix1337/Quasarr
└────────────────────────────────────┘""")

        port = int('8080')

        config_path = ""
        if os.environ.get('DOCKER'):
            config_path = "/config"
            if not arguments.internal_address:
                print(
                    "You must set the INTERNAL_ADDRESS variable to a locally reachable URL, e.g. http://192.168.1.1:8080")
                print("The local URL will be used by Radarr/Sonarr to connect to Quasarr")
                print("Stopping Quasarr...")
                sys.exit(1)
        else:
            if arguments.port:
                port = int(arguments.port)
            internal_address = f'http://{check_ip()}'

        if arguments.internal_address:
            internal_address = arguments.internal_address
        if arguments.external_address:
            external_address = arguments.external_address
        else:
            external_address = internal_address

        shared_state.set_connection_info(internal_address, external_address, port)

        if not config_path:
            config_path_file = "Quasarr.conf"
            if not os.path.exists(config_path_file):
                path_config(shared_state)
            with open(config_path_file, "r") as f:
                config_path = f.readline().strip()

        os.makedirs(config_path, exist_ok=True)

        try:
            temp_file = tempfile.TemporaryFile(dir=config_path)
            temp_file.close()
        except Exception as e:
            print(f'Could not access "{config_path}": {e}"'
                  f'Stopping Quasarr...')
            sys.exit(1)

        shared_state.set_files(config_path)
        shared_state.update("config", Config)
        shared_state.update("database", DataBase)
        shared_state.update("user_agent",
                            "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/131.0.0.0 Safari/537.36")
        shared_state.update("helper_active", False)

        print(f'Config path: "{config_path}"')

        shared_state.set_sites()

        try:
            if arguments.hostnames:
                hostnames_link = arguments.hostnames

                if is_valid_url(hostnames_link):
                    allowed_keys = extract_allowed_keys(Config._DEFAULT_CONFIG, 'Hostnames')
                    data = requests.get(hostnames_link).text
                    results = extract_kv_pairs(data, allowed_keys)

                    if results:
                        hostnames = Config('Hostnames')
                        for shorthand, hostname in results.items():
                            valid_domain = shared_state.extract_valid_hostname(hostname, shorthand)
                            if valid_domain:
                                hostnames.save(shorthand, hostname)
                                print(f'Hostname for "{shorthand}" extracted from {hostnames_link}.')
                    else:
                        print(f'No Hostnames found at "{hostnames_link}". '
                              'Ensure to pass a plain hostnames list, not html or json!')
                else:
                    print(f"Invalid hostnames URL: {hostnames_link}")
        except Exception as e:
            print(f'Error parsing hostnames link: {e}')

        if not get_clean_hostnames(shared_state):
            hostnames_config(shared_state)
            get_clean_hostnames(shared_state)

        if Config('Hostnames').get('nx'):
            nx = Config('Hostnames').get('nx')
            user = Config('NX').get('user')
            password = Config('NX').get('password')
            if not user or not password:
                nx_credentials_config(shared_state, nx)

        config = Config('JDownloader')
        user = config.get('user')
        password = config.get('password')
        device = config.get('device')

        if not user or not password or not device:
            jdownloader_config(shared_state)

        discord_url = ""
        if arguments.discord:
            discord_webhook_pattern = r'^https://discord\.com/api/webhooks/\d+/[\w-]+$'
            if re.match(discord_webhook_pattern, arguments.discord):
                shared_state.update("webhook", arguments.discord)
                print(f"Using Discord Webhook URL for CAPTCHA notifications")
                discord_url = arguments.discord
            else:
                print(f"Invalid Discord Webhook URL provided: {arguments.discord}")
        else:
            print("No Discord Webhook URL provided")
        shared_state.update("discord", discord_url)

        jdownloader = multiprocessing.Process(target=jdownloader_connection,
                                              args=(shared_state_dict, shared_state_lock))
        jdownloader.start()

        print(f'\nQuasarr API now running at "{shared_state.values["internal_address"]}"')
        print('Use this exact URL as "Newznab Indexer" and "SABnzbd Download Client" in Sonarr/Radarr')
        print("Leave settings at default and use this API key: 'quasarr'")

        protected = shared_state.get_db("protected").retrieve_all_titles()
        if protected:
            package_count = len(protected)
            print(f"\nCAPTCHA-Solution required for {package_count} package{'s' if package_count > 1 else ''} at "
                  f'{shared_state.values["external_address"]}/captcha"!\n')

        try:
            api(shared_state_dict, shared_state_lock)
        except KeyboardInterrupt:
            sys.exit(0)


def jdownloader_connection(shared_state_dict, shared_state_lock):
    shared_state.set_state(shared_state_dict, shared_state_lock)

    shared_state.set_device_from_config()

    connection_established = shared_state.get_device() and shared_state.get_device().name
    if not connection_established:
        i = 0
        while i < 10:
            i += 1
            print(f'Connection {i} to JDownloader failed. Device name: "{shared_state.values["device"]}"')
            time.sleep(60)
            shared_state.set_device_from_config()
            connection_established = shared_state.get_device() and shared_state.get_device().name
            if connection_established:
                break

    try:
        print(f'Connection to JDownloader successful. Device name: "{shared_state.get_device().name}"')
    except:
        print('Error connecting to JDownloader! Stopping Quasarr!')
        sys.exit(1)


class Unbuffered(object):
    def __init__(self, stream):
        self.stream = stream

    def write(self, data):
        self.stream.write(data)
        self.stream.flush()

    def writelines(self, datas):
        self.stream.writelines(datas)
        self.stream.flush()

    def __getattr__(self, attr):
        return getattr(self.stream, attr)


def check_ip():
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        s.connect(('10.255.255.255', 0))
        ip = s.getsockname()[0]
    except:
        ip = '127.0.0.1'
    finally:
        s.close()
    return ip


def is_valid_url(url):
    parsed = urlparse(url)
    return parsed.scheme in ("http", "https") and bool(parsed.netloc)


def extract_allowed_keys(config, section):
    """
    Extracts allowed keys from the specified section in the configuration.

    :param config: The configuration dictionary.
    :param section: The section from which to extract keys.
    :return: A list of allowed keys.
    """
    if section not in config:
        raise ValueError(f"Section '{section}' not found in configuration.")
    return [key for key, *_ in config[section]]


def extract_kv_pairs(input_text, allowed_keys):
    """
    Extracts key-value pairs from the given text where keys match allowed_keys.

    :param input_text: The input text containing key-value pairs.
    :param allowed_keys: A list of allowed two-letter shorthand keys.
    :return: A dictionary of extracted key-value pairs.
    """
    kv_pattern = re.compile(rf"^({'|'.join(map(re.escape, allowed_keys))})\s*=\s*(.*)$")
    kv_pairs = {}

    for line in input_text.splitlines():
        match = kv_pattern.match(line.strip())
        if match:
            key, value = match.groups()
            kv_pairs[key] = value

    return kv_pairs



# -*- coding: utf-8 -*-
# Quasarr
# Project by https://github.com/rix1337

import re
from base64 import urlsafe_b64encode

import requests
from bs4 import BeautifulSoup


def extract_size(text):
    match = re.match(r"(\d+)\s*([A-Za-z]+)", text)
    if match:
        size = match.group(1)
        unit = match.group(2)
        return {"size": size, "sizeunit": unit}
    else:
        raise ValueError(f"Invalid size format: {text}")


def fx_feed(shared_state):
    releases = []

    fx = shared_state.values["config"]("Hostnames").get("fx")

    password = fx.split(".")[0]
    url = f'https://{fx}/'
    headers = {
        'User-Agent': shared_state.values["user_agent"],
    }

    try:
        request = requests.get(url, headers=headers).content
        feed = BeautifulSoup(request, "html.parser")
        items = feed.findAll("article")
    except Exception as e:
        print(f"Error loading FX feed: {e}")
        return releases

    if items:
        for item in items:
            try:
                article = BeautifulSoup(str(item), "html.parser")
                try:
                    source = article.find('h2', class_='entry-title').a["href"]
                    titles = article.findAll("a", href=re.compile("(filecrypt|safe." + fx + ")"))
                except:
                    continue
                i = 0
                for title in titles:
                    link = title["href"]
                    title = (title.text.encode("ascii", errors="ignore").decode().
                             replace("/", "").replace(" ", ".").strip())

                    try:
                        size_info = article.findAll("strong", text=re.compile(r"(size|größe)", re.IGNORECASE))[
                            i].next.next.text.replace("|", "").strip()
                        size_item = extract_size(size_info)
                        mb = shared_state.convert_to_mb(size_item)
                        size = mb * 1024 * 1024
                        payload = urlsafe_b64encode(f"{title}|{link}|{mb}|{password}".encode("utf-8")).decode("utf-8")
                        link = f"{shared_state.values['internal_address']}/download/?payload={payload}"
                    except:
                        continue

                    try:
                        dates = article.findAll("time")
                        for date in dates:
                            published = date["datetime"]
                    except:
                        continue

                    releases.append({
                        "details": {
                            "title": f"[FX] {title}",
                            "link": link,
                            "size": size,
                            "date": published,
                            "source": source
                        },
                        "type": "protected"
                    })

            except Exception as e:
                print(f"Error parsing FX feed: {e}")

    return releases


def fx_search(shared_state, search_string):
    releases = []

    fx = shared_state.values["config"]("Hostnames").get("fx")

    password = fx.split(".")[0]
    url = f'https://{fx}/?s={search_string}'
    headers = {
        'User-Agent': shared_state.values["user_agent"],
    }

    try:
        request = requests.get(url, headers=headers).content
        search = BeautifulSoup(request, "html.parser")
        results = search.find('h2', class_='entry-title')

    except Exception as e:
        print(f"Error loading FX feed: {e}")
        return releases

    if results:
        for result in results:
            result_source = result["href"]
            try:
                request = requests.get(result_source, headers=headers).content
                feed = BeautifulSoup(request, "html.parser")
                items = feed.findAll("article")
            except Exception as e:
                print(f"Error loading FX feed: {e}")
                return releases

            for item in items:
                try:
                    article = BeautifulSoup(str(item), "html.parser")
                    try:
                        titles = article.findAll("a", href=re.compile("(filecrypt|safe." + fx + ")"))
                    except:
                        continue
                    i = 0
                    for title in titles:
                        link = title["href"]
                        title = (title.text.encode("ascii", errors="ignore").decode().
                                 replace("/", "").replace(" ", ".").strip())
                        try:
                            size_info = article.findAll("strong", text=re.compile(r"(size|größe)", re.IGNORECASE))[
                                i].next.next.text.replace("|", "").strip()
                            size_item = extract_size(size_info)
                            mb = shared_state.convert_to_mb(size_item)
                            size = mb * 1024 * 1024
                            payload = urlsafe_b64encode(f"{title}|{link}|{mb}|{password}".encode("utf-8")).decode(
                                "utf-8")
                            link = f"{shared_state.values['internal_address']}/download/?payload={payload}"
                        except:
                            continue

                        try:
                            dates = article.findAll("time")
                            for date in dates:
                                published = date["datetime"]
                        except:
                            continue

                        releases.append({
                            "details": {
                                "title": f"[FX] {title}",
                                "link": link,
                                "size": size,
                                "date": published,
                                "source": result_source
                            },
                            "type": "protected"
                        })

                except Exception as e:
                    print(f"Error parsing FX search: {e}")

    return releases



# -*- coding: utf-8 -*-
# Quasarr
# Project by https://github.com/rix1337

from concurrent.futures import ThreadPoolExecutor, as_completed

from quasarr.search.sources.dw import dw_feed, dw_search
from quasarr.search.sources.fx import fx_feed, fx_search
from quasarr.search.sources.nx import nx_feed, nx_search


def get_search_results(shared_state, request_from, search_string="", season="", episode=""):
    results = []

    dw = shared_state.values["config"]("Hostnames").get("dw")
    fx = shared_state.values["config"]("Hostnames").get("fx")
    nx = shared_state.values["config"]("Hostnames").get("nx")

    functions = []
    if search_string:
        if dw:
            functions.append(lambda: dw_search(shared_state, request_from, search_string))
        if fx:
            functions.append(lambda: fx_search(shared_state, search_string))
        if nx:
            functions.append(lambda: nx_search(shared_state, request_from, search_string))
    else:
        if dw:
            functions.append(lambda: dw_feed(shared_state, request_from))
        if fx:
            functions.append(lambda: fx_feed(shared_state))
        if nx:
            functions.append(lambda: nx_feed(shared_state, request_from))

    with ThreadPoolExecutor() as executor:
        futures = [executor.submit(func) for func in functions]
        for future in as_completed(futures):
            try:
                result = future.result()
                results.extend(result)
            except Exception as e:
                print(f"An error occurred: {e}")

    if search_string:
        print(f'Providing {len(results)} releases to {request_from} for search phrase "{search_string}"')
    else:
        print(f'Providing {len(results)} releases to {request_from} from release feed')

    return results



# -*- coding: utf-8 -*-
# Quasarr
# Project by https://github.com/rix1337

import json

from quasarr.downloads.sources.dw import get_dw_download_links
from quasarr.downloads.sources.nx import get_nx_download_links
from quasarr.providers.myjd_api import TokenExpiredException, RequestTimeoutException, MYJDException
from quasarr.providers.notifications import send_discord_message


def get_first_matching_comment(package, package_links):
    package_uuid = package.get("uuid")
    if package_uuid:
        for link in package_links:
            if link.get("packageUUID") == package_uuid:
                return link.get("comment")
    return None


def get_links_matching_package_uuid(package, package_links):
    package_uuid = package.get("uuid")
    link_ids = []
    if package_uuid:
        for link in package_links:
            if link.get("packageUUID") == package_uuid:
                link_ids.append(link.get("uuid"))
    return link_ids


def format_eta(seconds):
    hours = seconds // 3600
    minutes = (seconds % 3600) // 60
    seconds = seconds % 60
    return f"{hours:02}:{minutes:02}:{seconds:02}"


def get_packages(shared_state):
    packages = []

    protected_packages = shared_state.get_db("protected").retrieve_all_titles()
    if protected_packages:
        for package in protected_packages:
            package_id = package[0]

            data = json.loads(package[1])
            details = {
                "title": data["title"],
                "urls": data["links"],
                "size_mb": data["size_mb"],
                "password": data["password"]
            }

            packages.append({
                "details": details,
                "location": "queue",
                "type": "protected",
                "package_id": package_id
            })
    try:
        linkgrabber_packages = shared_state.get_device().linkgrabber.query_packages()
    except (TokenExpiredException, RequestTimeoutException, MYJDException):
        linkgrabber_packages = []

    if linkgrabber_packages:
        for package in linkgrabber_packages:
            comment = get_first_matching_comment(package, shared_state.get_device().linkgrabber.query_links())
            packages.append({
                "details": package,
                "location": "queue",
                "type": "linkgrabber",
                "comment": comment,
                "uuid": package.get("uuid")
            })
    try:
        downloader_packages = shared_state.get_device().downloads.query_packages()
    except (TokenExpiredException, RequestTimeoutException, MYJDException):
        downloader_packages = []

    if downloader_packages:
        for package in downloader_packages:
            comment = get_first_matching_comment(package, shared_state.get_device().downloads.query_links())
            status = package.get("status", "")

            if any(ex_str in status.lower() for ex_str in ["entpacken", "extracting"]) and "ok:" not in status.lower():
                finished = False
            else:
                finished = package.get("finished", False)

            packages.append({
                "details": package,
                "location": "history" if finished else "queue",
                "type": "downloader",
                "comment": comment,
                "uuid": package.get("uuid")
            })

    downloads = {
        "queue": [],
        "history": []
    }
    for package in packages:
        queue_index = 0
        history_index = 0

        if package["location"] == "queue":
            time_left = "2376:00:00"  # yields "99d" to signify that its not running
            if package["type"] == "linkgrabber":
                details = package["details"]
                name = f"[Linkgrabber] {details["name"]}"
                try:
                    mb = mb_left = int(details["bytesTotal"]) / (1024 * 1024)
                except KeyError:
                    mb = mb_left = 0
                try:
                    package_id = package["comment"]
                    if "movies" in package_id:
                        category = "movies"
                    else:
                        category = "tv"
                except TypeError:
                    category = "not_quasarr"
                package_type = "linkgrabber"
                package_uuid = package["uuid"]
            elif package["type"] == "downloader":
                details = package["details"]
                name = f"[Downloading] {details["name"]}"
                try:
                    if details["eta"]:
                        time_left = format_eta(int(details["eta"]))
                except KeyError:
                    name = name.replace("[Downloading]", "[Paused]")
                try:
                    mb = int(details["bytesTotal"]) / (1024 * 1024)
                    mb_left = (int(details["bytesTotal"]) - int(details["bytesLoaded"])) / (1024 * 1024)
                except KeyError:
                    mb = mb_left = 0
                try:
                    package_id = package["comment"]
                    if "movies" in package_id:
                        category = "movies"
                    else:
                        category = "tv"
                except TypeError:
                    category = "not_quasarr"
                package_type = "downloader"
                package_uuid = package["uuid"]
            else:
                details = package["details"]
                name = f"[CAPTCHA not solved!] {details["title"]}"
                mb = mb_left = details["size_mb"]
                try:
                    package_id = package["package_id"]
                    if "movies" in package_id:
                        category = "movies"
                    else:
                        category = "tv"
                except TypeError:
                    category = "not_quasarr"
                package_type = "protected"
                package_uuid = None

            try:
                downloads["queue"].append({
                    "index": queue_index,
                    "nzo_id": package_id,
                    "priority": "Normal",
                    "filename": name,
                    "cat": category,
                    "mbleft": int(mb_left),
                    "mb": int(mb),
                    "status": "Downloading",
                    "timeleft": time_left,
                    "type": package_type,
                    "uuid": package_uuid
                })
            except:
                print(f"Parameters missing for {package}")
            queue_index += 1
        elif package["location"] == "history":
            details = package["details"]
            name = details["name"]
            size = int(details["bytesLoaded"])
            storage = details["saveTo"]
            try:
                package_id = package["comment"]
                if "movies" in package_id:
                    category = "movies"
                else:
                    category = "tv"
            except TypeError:
                category = "not_quasarr"

            downloads["history"].append({
                "fail_message": "",
                "category": category,
                "storage": storage,
                "status": "Completed",
                "nzo_id": package_id,
                "name": name,
                "bytes": int(size),
                "type": "downloader",
                "uuid": package["uuid"]
            })
            history_index += 1
        else:
            print(f"Invalid package location {package['location']}")

    return downloads


def delete_package(shared_state, package_id):
    deleted = ""

    packages = get_packages(shared_state)
    for package_location in packages:
        for package in packages[package_location]:
            if package["nzo_id"] == package_id:
                if package["type"] == "linkgrabber":
                    ids = get_links_matching_package_uuid(package, shared_state.get_device().linkgrabber.query_links())
                    shared_state.get_device().linkgrabber.remove_links(ids, [package["uuid"]])
                elif package["type"] == "downloader":
                    ids = get_links_matching_package_uuid(package, shared_state.get_device().downloads.query_links())
                    shared_state.get_device().downloads.remove_links(ids, [package["uuid"]])
                else:
                    shared_state.get_db("protected").delete(package_id)
                if package_location == "queue":
                    package_name_field = "filename"
                else:
                    package_name_field = "name"

                deleted = package[package_name_field]
                break
        if deleted:
            break

    if deleted:
        print(f"Deleted package {deleted} with ID {package_id}")
    else:
        print(f"Failed to delete package {package_id}")
    return deleted


def download_package(shared_state, request_from, title, url, size_mb, password):
    if "radarr".lower() in request_from.lower():
        category = "movies"
    else:
        category = "tv"

    package_id = f"Quasarr_{category}_{str(hash(title + url)).replace('-', '')}"

    dw = shared_state.values["config"]("Hostnames").get("dw")
    nx = shared_state.values["config"]("Hostnames").get("nx")

    if nx and nx.lower() in url.lower():
        links = get_nx_download_links(shared_state, url, title)
        if links:
            print(f"Decrypted {len(links)} download links for {title}")
            send_discord_message(shared_state, title=title, case="unprotected")
            added = shared_state.download_package(links, title, password, package_id)
            if not added:
                print(f"Failed to add {title} to linkgrabber")
                package_id = None
        else:
            print(f"Found 0 links decrypting {title}")
            package_id = None

    elif dw and dw.lower() in url.lower():
        links = get_dw_download_links(shared_state, url, title)
        print(f"CAPTCHA-Solution required for {title} - {shared_state.values['external_address']}/captcha")
        send_discord_message(shared_state, title=title, case="captcha")
        blob = json.dumps({"title": title, "links": links, "size_mb": size_mb, "password": password})
        shared_state.values["database"]("protected").update_store(package_id, blob)

    elif "filecrypt".lower() in url.lower():
        print(f"CAPTCHA-Solution required for {title} - {shared_state.values['external_address']}/captcha")
        send_discord_message(shared_state, title=title, case="captcha")
        blob = json.dumps({"title": title, "links": [[url, "filecrypt"]], "size_mb": size_mb, "password": password})
        shared_state.values["database"]("protected").update_store(package_id, blob)

    else:
        package_id = None
        print(f"Could not parse URL for {title} - {url}")

    return package_id



# -*- coding: utf-8 -*-
# Quasarr
# Project by https://github.com/rix1337

import os
import time
from urllib import parse

from quasarr.providers.myjd_api import Myjdapi, TokenExpiredException, RequestTimeoutException, MYJDException, Jddevice
from quasarr.storage.config import Config
from quasarr.storage.sqlite_database import DataBase

values = {}
lock = None


def set_state(manager_dict, manager_lock):
    global values
    global lock
    values = manager_dict
    lock = manager_lock


def update(key, value):
    global values
    global lock
    lock.acquire()
    try:
        values[key] = value
    finally:
        lock.release()


def set_sites():
    update("sites", ["DW", "FX", "NX"])


def set_connection_info(internal_address, external_address, port):
    if internal_address.count(":") < 2:
        internal_address = f"{internal_address}:{port}"
    update("internal_address", internal_address)
    update("external_address", external_address)
    update("port", port)


def set_files(config_path):
    update("configfile", os.path.join(config_path, "Quasarr.ini"))
    update("dbfile", os.path.join(config_path, "Quasarr.db"))


def connect_to_jd(jd, user, password, device_name):
    try:
        jd.connect(user, password)
        jd.update_devices()
        device = jd.get_device(device_name)
    except (TokenExpiredException, RequestTimeoutException, MYJDException) as e:
        print("Error connecting to JDownloader: " + str(e))
        return False
    if not device or not isinstance(device, (type, Jddevice)):
        return False
    else:
        device.downloadcontroller.get_current_state()  # request forces direct_connection info update
        connection_info = device.check_direct_connection()
        if connection_info["status"]:
            print(f'Direct connection to JDownloader established: "{connection_info['ip']}"')
        else:
            print("Could not establish direct connection to JDownloader.")
        update("device", device)
        return True


def set_device(user, password, device):
    jd = Myjdapi()
    jd.set_app_key('Quasarr')
    return connect_to_jd(jd, user, password, device)


def set_device_from_config():
    config = Config('JDownloader')
    user = str(config.get('user'))
    password = str(config.get('password'))
    device = str(config.get('device'))

    update("device", device)

    if user and password and device:
        jd = Myjdapi()
        jd.set_app_key('Quasarr')
        return connect_to_jd(jd, user, password, device)
    return False


def check_device(device):
    try:
        valid = isinstance(device,
                           (type, Jddevice)) and device.downloadcontroller.get_current_state()
    except (AttributeError, KeyError, TokenExpiredException, RequestTimeoutException, MYJDException):
        valid = False
    return valid


def connect_device():
    config = Config('JDownloader')
    user = str(config.get('user'))
    password = str(config.get('password'))
    device = str(config.get('device'))

    jd = Myjdapi()
    jd.set_app_key('Quasarr')

    if user and password and device:
        try:
            jd.connect(user, password)
            jd.update_devices()
            device = jd.get_device(device)
        except (TokenExpiredException, RequestTimeoutException, MYJDException):
            pass

    if check_device(device):
        update("device", device)
        return True
    else:
        return False


def get_device():
    attempts = 0

    while True:
        try:
            if check_device(values["device"]):
                break
        except (AttributeError, KeyError, TokenExpiredException, RequestTimeoutException, MYJDException):
            pass
        attempts += 1

        update("device", False)

        if attempts % 10 == 0:
            print(
                f"WARNING: {attempts} consecutive JDownloader connection errors. Please check your credentials!")
        time.sleep(3)

        if connect_device():
            break

    return values["device"]


def get_devices(user, password):
    jd = Myjdapi()
    jd.set_app_key('Quasarr')
    try:
        jd.connect(user, password)
        jd.update_devices()
        devices = jd.list_devices()
        return devices
    except (TokenExpiredException, RequestTimeoutException, MYJDException) as e:
        print("Error connecting to JDownloader: " + str(e))
        return []


def get_db(table):
    return DataBase(table)


def convert_to_mb(item):
    size = float(item['size'])
    unit = item['sizeunit'].upper()

    if unit == 'B':
        size_b = size
    elif unit == 'KB':
        size_b = size * 1024
    elif unit == 'MB':
        size_b = size * 1024 * 1024
    elif unit == 'GB':
        size_b = size * 1024 * 1024 * 1024
    elif unit == 'TB':
        size_b = size * 1024 * 1024 * 1024 * 1024
    else:
        raise ValueError(f"Unsupported size unit {item['name']} {item['size']} {item['sizeunit']}")

    size_mb = size_b / (1024 * 1024)
    return int(size_mb)


def download_package(links, title, password, package_id):
    device = get_device()
    device.linkgrabber.add_links(params=[
        {
            "autostart": False,
            "links": str(links).replace(" ", ""),
            "packageName": title,
            "extractPassword": password,
            "priority": "DEFAULT",
            "downloadPassword": password,
            "destinationFolder": "Quasarr/<jd:packagename>",
            "comment": package_id,
            "overwritePackagizerRules": True
        }
    ])

    package_uuids = []
    link_ids = []
    archive_id = None

    for _ in range(30):
        try:
            collecting = device.linkgrabber.is_collecting()
            if not collecting:
                links = device.linkgrabber.query_links()
                for link in links:
                    if link["comment"] == package_id:
                        link_id = link["uuid"]
                        if link_id not in link_ids:
                            link_ids.append(link_id)
                        package_uuid = link["packageUUID"]
                        if package_uuid not in package_uuids:
                            package_uuids.append(package_uuid)

                if link_ids and package_uuids:
                    archive = device.extraction.get_archive_info(link_ids=link_ids, package_ids=package_uuids)
                    if archive:
                        archive_id = archive[0].get("archiveId", None)
                        if archive_id:
                            break  # Exit the loop as archive_id is found

        except Exception as e:
            print(f"An error occurred: {e}")

        time.sleep(1)

    if not link_ids and not package_uuids:
        print(f"No links or packages found within 30 seconds! Adding {title} package failed.")
        return False

    if not archive_id:
        print(f"Archive ID for {title} not found! Release may not be compressed.")
    else:
        settings = {
            "autoExtract": True,
            "removeDownloadLinksAfterExtraction": False,
            "removeFilesAfterExtraction": True
        }
        settings_set = device.extraction.set_archive_settings(archive_id, archive_settings=settings)
        if not settings_set:
            print(f"Failed to set archive settings for {title}!")

    time.sleep(3)
    links = device.linkgrabber.query_links()
    for link in links:
        if link["comment"] == package_id:
            link_id = link["uuid"]
            if link_id not in link_ids:
                link_ids.append(link_id)
            package_uuid = link["packageUUID"]
            if package_uuid not in package_uuids:
                package_uuids.append(package_uuid)

    downloaded = False

    try:
        for _ in range(30):
            collecting = device.linkgrabber.is_collecting()
            if not collecting:
                device.linkgrabber.move_to_downloadlist(link_ids, package_uuids)
                downloaded = True
            time.sleep(1)
    except Exception as e:
        print(f"Failed to start download for {title}: {e}")
        return False

    if not downloaded:
        print(f"Failed to start download for {title}: Linkgrabber timeout!")
    return downloaded


def extract_valid_hostname(url, shorthand):
    # Check if both characters from the shorthand appear in the url
    try:
        if '://' not in url:
            url = 'http://' + url
        result = parse.urlparse(url)
        domain = result.netloc

        # Check if both characters in the shorthand are in the domain
        if all(char in domain for char in shorthand):
            print(f"{domain} matches both characters from {shorthand}. Continuing...")
            return domain
        else:
            print(f"Invalid domain {domain}: Does not contain both characters from shorthand {shorthand}")
            return None
    except Exception as e:
        print(f"Error parsing URL {url}: {e}")
        return None



# -*- coding: utf-8 -*-
# Quasarr
# Project by https://github.com/rix1337

import re

import requests


def get_localized_title(shared_state, imdb_id, language='de'):
    localized_title = None

    headers = {
        'Accept-Language': language,
        'User-Agent': shared_state.values["user_agent"]
    }

    try:
        response = requests.get(f"https://www.imdb.com/title/{imdb_id}/", headers=headers)
    except Exception as e:
        print(f"Error loading IMDb metadata for {imdb_id}: {e}")
        return localized_title

    try:
        match = re.findall(r'<title>(.*?) \(.*?</title>', response.text)
        localized_title = match[0]
    except:
        try:
            match = re.findall(r'<title>(.*?) - IMDb</title>', response.text)
            localized_title = match[0]
        except:
            pass

    if not localized_title:
        print(f"Could not get localized title for {imdb_id} in {language} from IMDb")

    return localized_title



# -*- coding: utf-8 -*-
# Quasarr
# Project by https://github.com/rix1337

import json

import requests


def send_discord_message(shared_state, title, case):
    """
    Sends a Discord message to the webhook provided in the shared state, based on the specified case.

    :param shared_state: Shared state object containing configuration.
    :param title: Title of the embed to be sent.
    :param case: A string representing the scenario (e.g., 'captcha', 'captcha_solved', 'package_deleted').
    :return: True if the message was sent successfully, False otherwise.
    """
    if not shared_state.values.get("discord"):
        return False

    # Decide the embed content based on the case
    if case == "unprotected":
        description = 'No CAPTCHA required. Links were added directly!'
        fields = None
    elif case == "solved":
        description = 'CAPTCHA solved by SponsorsHelper!'
        fields = None
    elif case == "deleted":
        description = 'SponsorsHelper failed to solve the CAPTCHA! Package deleted.'
        fields = None
    elif case == "captcha":
        if shared_state.values.get("helper_active"):
            helper_text = f"Just wait - SponsorsHelper will solve this CAPTCHA for you shortly."
        else:
            helper_text = f'[Become a Sponsor and let SponsorsHelper solve CAPTCHAs for you!]({f"https://github.com/users/rix1337/sponsorship"})'

        description = 'Links are protected by a CAPTCHA! Chose how to proceed:'
        fields = [
            {
                'name': 'Automatically',
                'value': helper_text,
            },
            {
                'name': 'Manually',
                'value': f'[Solve the CAPTCHA here to start the download.]({f"{shared_state.values['external_address']}/captcha"})',
            }
        ]
    else:
        print(f"Unknown notification case: {case}")
        return False

    data = {
        'username': 'Quasarr',
        'avatar_url': 'https://raw.githubusercontent.com/rix1337/Quasarr/main/Quasarr.png',
        'embeds': [{
            'title': title,
            'description': description,
        }]
    }

    if fields:
        data['embeds'][0]['fields'] = fields

    response = requests.post(shared_state.values["discord"], data=json.dumps(data),
                             headers={"Content-Type": "application/json"})
    if response.status_code != 204:
        print(f"Failed to send message to Discord webhook. Status code: {response.status_code}")
        return False

    return True



# -*- coding: utf-8 -*-
# Quasarr
# Project by https://github.com/rix1337

import html
import re
from base64 import urlsafe_b64encode

import requests

from quasarr.providers.imdb_metadata import get_localized_title


def nx_feed(shared_state, request_from):
    releases = []
    nx = shared_state.values["config"]("Hostnames").get("nx")
    password = nx

    if "Radarr" in request_from:
        category = "movie"
    else:
        category = "episode"

    url = f'https://{nx}/api/frontend/releases/category/{category}/tag/all/1/51?sort=date'
    headers = {
        'User-Agent': shared_state.values["user_agent"],
    }

    try:
        response = requests.get(url, headers)
        feed = response.json()
    except Exception as e:
        print(f"Error loading NX feed: {e}")
        return releases

    items = feed['result']['list']
    for item in items:
        try:
            title = item['name']
            if title:
                try:
                    source = f"https://{nx}/release/{item['slug']}"
                    mb = shared_state.convert_to_mb(item)
                    payload = urlsafe_b64encode(f"{title}|{source}|{mb}|{password}".encode("utf-8")).decode("utf-8")
                    link = f"{shared_state.values['internal_address']}/download/?payload={payload}"
                except:
                    continue

                try:
                    size = mb * 1024 * 1024
                except:
                    continue

                try:
                    published = item['publishat']
                except:
                    continue

                releases.append({
                    "details": {
                        "title": f"[NX] {title}",
                        "link": link,
                        "size": size,
                        "date": published,
                        "source": source
                    },
                    "type": "protected"
                })

        except Exception as e:
            print(f"Error parsing NX feed: {e}")

    return releases


def nx_search(shared_state, request_from, search_string):
    releases = []
    nx = shared_state.values["config"]("Hostnames").get("nx")
    password = nx

    if "Radarr" in request_from:
        valid_type = "movie"
    else:
        valid_type = "episode"

    if re.match(r'^tt\d{7,8}$', search_string):
        imdb_id = search_string
        search_string = get_localized_title(shared_state, imdb_id, 'de')
        if not search_string:
            print(f"Could not extract title from IMDb-ID {imdb_id}")
            return releases
        search_string = html.unescape(search_string)

    url = f'https://{nx}/api/frontend/search/{search_string}'
    headers = {
        'User-Agent': shared_state.values["user_agent"],
    }

    try:
        response = requests.get(url, headers)
        feed = response.json()
    except Exception as e:
        print(f"Error loading NX search: {e}")
        return releases

    items = feed['result']['releases']
    for item in items:
        try:
            if item['type'] == valid_type:
                title = item['name']
                if title:
                    try:
                        source = f"https://{nx}/release/{item['slug']}"
                        mb = shared_state.convert_to_mb(item)
                        payload = urlsafe_b64encode(f"{title}|{source}|{mb}|{password}".
                                                    encode("utf-8")).decode("utf-8")
                        link = f"{shared_state.values['internal_address']}/download/?payload={payload}"
                    except:
                        continue

                    try:
                        size = mb * 1024 * 1024
                    except:
                        continue

                    try:
                        published = item['publishat']
                    except:
                        published = ""

                    releases.append({
                        "details": {
                            "title": f"[NX] {title}",
                            "link": link,
                            "size": size,
                            "date": published,
                            "source": source
                        },
                        "type": "protected"
                    })

        except Exception as e:
            print(f"Error parsing NX search: {e}")

    return releases
